@@ -58,20 +58,6 @@ def test_open_write(temp, fs):
         assert f.read() == content
 
 
-def test_store_access(temp, fs):
-    paths = []
-    with fs.store(temp, ["ex1.txt", "ex2.txt"]) as datafiles:
-        for d in datafiles:
-            paths.append(d.path)
-            with d.handle("wb") as f:
-                f.write(TEST_STRING)
-
-    datafiles = fs.access(paths)
-    for d in datafiles:
-        with d.handle("rb") as f:
-            assert f.read() == TEST_STRING
-
-
 def test_copy_recursive_to_local(populated, tmpdir, fs):
     dest = str(tmpdir)
     fs.cp(populated, dest, recursive=True)
@@ -82,10 +68,13 @@ def test_copy_recursive_to_local(populated, tmpdir, fs):
 
 def test_copy_recursive_matched(populated, fs):
     dest = populated.replace("data", "copy")
-    fs.cp(populated, dest, recursive=True)
-    source = [p.replace(populated, "") for p in fs.ls(populated + "/**")]
-    copy = fs.ls(dest + "/**")
-    assert (s in c for s, c in zip(source, copy))
+    try:
+        fs.cp(populated, dest, recursive=True)
+        source = [p.replace(populated, "") for p in fs.ls(populated + "/**")]
+        copy = fs.ls(dest + "/**")
+        assert (s in c for s, c in zip(source, copy))
+    finally:
+        fs.rm(dest, recursive=True)
 
 
 def test_rm(populated, fs):
